/* Copyright 1999 by Herve Regad-Pellagru, E-mail: regad@micronet.fr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <unistd.h>

#include <stdio.h>
#include <sys/fcntl.h>
#include <sys/stat.h>

#include <termios.h>
#include <sys/ioctl.h>


#include "cm12a.h"
#include "proto.h"
#include "error.h"


/* debug stuff */
#ifdef DEBUG 
#define P_DEBUG1(fmt,arg) {fprintf(stderr,fmt,arg);}
#define P_DEBUG2(fmt,arg1,arg2) {fprintf(stderr,fmt,arg1,arg2);}
#else
#define P_DEBUG1(fmt,arg)
#define P_DEBUG2(fmt,arg1,arg2)
#endif


/* x10_open_tty_port() opens the serial port in an CM11A-friendly
   way (8 bits, 4800 bps, no parity) */
int x10_open_tty_port(char *filename)
{
  struct termios oldsb, newsb;
  int fd;
  int i;

  if((fd = open(filename,O_RDWR | O_NOCTTY)) == -1)
    return(OPEN_FAILED);

  /* get the attributes. Not sure if this is POSIX ... */
  if (tcgetattr(fd, &oldsb) < 0)
    return(TCGETATTR_FAILED);

  newsb = oldsb;
  newsb.c_iflag = IGNBRK | IGNPAR;
  newsb.c_oflag = 0;
  newsb.c_lflag = ISIG;
  newsb.c_cflag = (CLOCAL | B4800 | CS8 | CREAD);

  for (i = 0; i < NCCS; i++)
    newsb.c_cc[i] = 0;

  newsb.c_cc[VMIN] = 1;
  newsb.c_cc[VTIME] = 0;

  /* set attributes after setting them, immediate effect */
  tcsetattr(fd, TCSANOW, &newsb);
  return(fd);
}


/* Just close the port */
int x10_close_tty_port(int fd)
{
  return(close(fd));
}


/* Well, to read from the CM11A is a bit tricky as the damn thing can poll the PC
   and won't do anything until the poll is satisfied. Also, to avoid blocking, we
   use select with a timeout to actually get a chance of doing something else
   if nothing happens on the AC bus. Also, x10_read() has a flag called check_poll
   which tells whether we should consider polling codes as exception or not. */
int x10_read(struct x10_status_t *status,unsigned char *buffer,int leng,int useconds,
	     int check_poll)
{
  struct timeval time_wait;
  int return_read,real_len;
  int return_complete;
  int total_read;

  /* fill time_wait struct for select(2), and set the FD_SET structure */
  time_wait.tv_sec = useconds/1000000;
  time_wait.tv_usec = useconds % 1000000;
  FD_ZERO(&(status->serial_set));
  FD_SET(status->fdinout,&(status->serial_set));

  real_len = select(status->fdinout+1,&(status->serial_set),NULL,NULL,&time_wait);

  P_DEBUG1("x10_read: select returns %d\n",real_len);

  switch(real_len)
    {
      /* select has failed -> error */
    case -1:
      P_DEBUG1("x10_read: %s FAILED\n","select");
      return(SELECT_FAILED);

    case 0:
      /* nothing to read after timeout */
      P_DEBUG1("x10_read: TIMEOUT in %s\n","read");
      return(TIMEOUT_READ);

    default:
      /* something to read */
      break;
    }

  /* Read only one byte, as it may be a poll and we have to treat it before
     coming back to our business */
  return_read=read(status->fdinout,buffer,1);

  P_DEBUG2("x10_read: reading %d bytes out of %d bytes\n",return_read,leng);

  if(return_read<=0)
    {
      P_DEBUG1("x10_read: %s FAILED\n","read");
      return(READ_FAILED);
    }

  if(check_poll)
    switch(buffer[0])
      {
	/* Is it a poll ? */
      case X10_INTERFACE_POLL:
	/* yes, we get the data, evaluate and decode it, and trigger whatever
	   event handler has been specified */
	P_DEBUG1("x10_read: receiving %x as INTERFACE_POLL\n",buffer[0]);
	return_complete=x10_complete_events(status);
	if(return_complete < 0)
	  return(return_complete);

	if(!return_complete)
	  return(x10_read(status,buffer,leng,useconds,check_poll));

	return(CHAT_INTERRUPTED);
	break;

      /* Did we get a power fail ? */
      case X10_POWER_FAIL:
	/* We set the date and time */
	P_DEBUG1("x10_read: receiving %x as POWER_FAIL\n",buffer[0]);
	return_complete=x10_set_status_interface(status);
	return(return_complete ? return_complete : CHAT_INTERRUPTED);
	break;
      }

  /* This wasn't a poll: we read the remaining bytes. Indeed, there's no poll
     to take care of, here, just the waited bytes */
  for(total_read=1;total_read<leng;total_read++)
    if(read(status->fdinout,buffer+total_read,1) != 1)
      return(READ_FAILED);

  P_DEBUG2("x10_read: reading remaining %d bytes out of %d bytes\n",leng-1,leng);

  return(total_read);
}


/* X10_write is quite trivial */
int x10_write(struct x10_status_t *status,unsigned char *buffer,int leng)
{
  return(write(status->fdinout,buffer,leng));
}
